@@ -9,8 +9,10 @@ use sphinx_lang::parser::expr::Expr;
 use sphinx_lang::parser::stmt::{Stmt, StmtMeta};
 use sphinx_lang::codegen::{Program, CompiledProgram};
 use sphinx_lang::runtime::VirtualMachine;
-use sphinx_lang::runtime::module::{ModuleCache, GlobalEnv};
+use sphinx_lang::runtime::gc::GC;
+use sphinx_lang::runtime::module::{Module, GlobalEnv};
 use sphinx_lang::runtime::strings::StringInterner;
+use sphinx_lang::stdlib::prelude;
 use sphinx_lang::debug::symbol::resolver::BufferedResolver;
 
 fn main() {
@@ -65,11 +67,11 @@ fn main() {
         source = ModuleSource::File(PathBuf::from(s));
         name = s;
     } else {
-        let mut module_cache = ModuleCache::new();
-        let repl_env = GlobalEnv::new();
+        let repl_env = GlobalEnv::allocate();
+        repl_env.borrow_mut().extend(&prelude::create_prelude());
+        
+        Repl::new(version.to_string(), repl_env).run();
         
-        println!("\nSphinx Version {}\n", version);
-        Repl::new(&mut module_cache, &repl_env).run();
         return;
     }
     
@@ -83,26 +85,30 @@ fn main() {
         if let Some(build) = build_program(&args, name, &source) {
             let program = Program::load(build.program);
             
-            let mut module_cache = ModuleCache::new();
-            let module_id = module_cache.insert(program.data, Some(name.to_string()), Some(source));
+            let repl_env = GlobalEnv::allocate();
+            repl_env.borrow_mut().extend(&prelude::create_prelude());
             
+            let main_module = Module::with_globals(repl_env, Some(source), program.data);
             
-            let vm = VirtualMachine::new(&module_cache, module_id, &program.main);
-            vm.run().expect("runtime error");
+            let vm = VirtualMachine::new(main_module, &program.main);
+            if args.is_present("debug") {
+                run_debugger(vm);
+            } else {
+                vm.run().expect("runtime error");
+            }
             
-            println!("\nSphinx Version {}\n", version);
-            
-            let repl_env = module_cache.get(&module_id).unwrap().globals().clone();
-            Repl::new(&mut module_cache, &repl_env).run()
+            Repl::new(version.to_string(), repl_env).run()
         }
     }
     else if let Some(build) = build_program(&args, name, &source) {
         let program = Program::load(build.program);
         
-        let mut module_cache = ModuleCache::new();
-        let module_id = module_cache.insert(program.data, Some(name.to_string()), Some(source));
+        let main_env = GlobalEnv::allocate();
+        main_env.borrow_mut().extend(&prelude::create_prelude());
         
-        let vm = VirtualMachine::new(&module_cache, module_id, &program.main);
+        let main_module = Module::with_globals(main_env, Some(source), program.data);
+        
+        let vm = VirtualMachine::new(main_module, &program.main);
         if args.is_present("debug") {
             run_debugger(vm);
         } else {
@@ -188,9 +194,9 @@ fn parse_and_print_ast(_args: &ArgMatches, name: &str, source: &ModuleSource) {
 const PROMT_START: &str = ">>> ";
 const PROMT_CONTINUE: &str = "... ";
 
-struct Repl<'m> {
-    module_cache: &'m mut ModuleCache,
-    repl_env: &'m GlobalEnv,
+struct Repl {
+    version: String,
+    repl_env: GC<GlobalEnv>,
 }
 
 enum ReadLine {
@@ -200,10 +206,10 @@ enum ReadLine {
     Quit,
 }
 
-impl<'m> Repl<'m> {
-    pub fn new(module_cache: &'m mut ModuleCache, repl_env: &'m GlobalEnv) -> Self {
+impl Repl {
+    pub fn new(version: String, repl_env: GC<GlobalEnv>) -> Self {
         Self {
-            module_cache, repl_env,
+            version, repl_env,
         }
     }
     
@@ -232,6 +238,7 @@ impl<'m> Repl<'m> {
     }
     
     pub fn run(&mut self) {
+        println!("\nSphinx Version {}\n", self.version);
         
         loop {
             let mut interner;
@@ -309,17 +316,13 @@ impl<'m> Repl<'m> {
             
             let program = Program::load(build.program);
             
-            let module_id = self.module_cache.insert_with_globals(program.data, None, None, self.repl_env.clone());
+            let module = Module::with_globals(self.repl_env, None, program.data);
             
-            let vm = VirtualMachine::repl(self.module_cache, self.repl_env, module_id, &program.main);
+            let vm = VirtualMachine::new(module, &program.main);
             if let Err(error) = vm.run() {
                 println!("Runtime error: {:?}", error);
             }
             
-            // This is super inefficient, but it's the REPL, so that's okay?
-            let module = self.module_cache.get(&module_id).unwrap();
-            self.repl_env.borrow_mut().extend(&module.globals().borrow());
-            
         }
         
     }
