@@ -1,11 +1,12 @@
-use crate::runtime::module::{ModuleID, ChunkID};
+use crate::runtime::gc::GC;
+use crate::runtime::module::{Module, ChunkID};
 
 /// Traceback information
 #[derive(Debug, Clone)]
 pub enum CallSite {
     Chunk {
         offset: usize,
-        module_id: ModuleID,
+        module: GC<Module>,
         chunk_id: Option<ChunkID>,
     },
     Native,
