@@ -1,4 +1,4 @@
-use crate::runtime::strings::InternSymbol;
+use crate::language::InternSymbol;
 use crate::parser::lvalue::DeclType;
 use crate::parser::expr::{ExprMeta, ExprBlock};
 use crate::parser::stmt::StmtMeta;
