@@ -3,30 +3,48 @@ use std::rc::Rc;
 use std::hash::{Hash, Hasher};
 use std::cmp::{PartialEq, Eq};
 use crate::language::{IntType, FloatType};
-use crate::runtime::types::{Metatable, Call};
+use crate::runtime::types::Type;
+use crate::runtime::types::metatable::Metatable;
+use crate::runtime::types::function::{Function, NativeFunction, Call, Invoke};
 use crate::runtime::types::primitive::*;
 use crate::runtime::strings::{StringSymbol, STRING_TABLE};
-use crate::runtime::gc::{GCHandle, GCObject};
+use crate::runtime::gc::{GC, GCArray, GCTrace};
 use crate::runtime::errors::{ExecResult, RuntimeError, ErrorKind};
 
 
 // Fundamental data value type
-#[derive(Clone)] // add Copy?
+#[derive(Clone, Copy)]
 pub enum Variant {
     Nil,
     EmptyTuple,
     BoolTrue,
     BoolFalse,
+    
     Integer(IntType),
     Float(FloatType),
     String(StringSymbol),
-    // TODO just GC tuples
-    Tuple(Rc<[Variant]>),  //  will use COW semantics, so if we need to send to another thread we can just clone the underlying data
-    Object(GCHandle),
-    // LoadedModule(ModuleID)
+    
+    Tuple(GCArray<Variant>),
+    Function(GC<Function>),
+    NativeFunction(GC<NativeFunction>),
 }
 
 impl Variant {
+    pub fn get_type(&self) -> Type {
+        match self {
+            Self::Nil => Type::Nil,
+            Self::BoolTrue => Type::Boolean,
+            Self::BoolFalse => Type::Boolean,
+            Self::Integer(..) => Type::Integer,
+            Self::Float(..) => Type::Float,
+            Self::String(..) => Type::String,
+            Self::EmptyTuple => Type::Tuple,
+            Self::Tuple(..) => Type::Tuple,
+            Self::Function(..) => Type::Function,
+            Self::NativeFunction(..) => Type::Function,
+        }
+    }
+    
     pub fn metatable(&self) -> &Metatable {
         match self {
             Self::String(..) => &METATABLE_STRING,
@@ -59,10 +77,6 @@ impl Variant {
         Some(value)
     }
     
-    pub fn make_tuple(items: Box<[Variant]>) -> Self {
-        Self::Tuple(Rc::from(items))
-    }
-    
     pub fn can_hash(&self) -> bool {
         match self {
             Self::Float(..) => false,
@@ -74,16 +88,12 @@ impl Variant {
     }
     
     pub fn invoke(&self, args: &[Variant]) -> ExecResult<Call> {
-        // TODO make GC objects suck less
-        if let Self::Object(handle) = self {
-            return handle.with_ref(|obj| {
-                let GCObject::Function(fun) = &*obj;
-                fun.signature().check_args(args)?;
-                Ok(fun.as_call())
-            })
+        match self {
+            Self::Function(fun) => fun.invoke(args),
+            Self::NativeFunction(fun) => fun.invoke(args),
+            
+            _ => Err(ErrorKind::NotCallable(self.clone()).into())
         }
-        
-        Err(ErrorKind::NotCallable(self.clone()).into())
     }
 }
 
@@ -109,16 +119,31 @@ impl From<StringSymbol> for Variant {
 }
 
 impl From<&str> for Variant {
-    fn from(value: &str) -> Self {
-        let symbol = STRING_TABLE.with(|string_table| string_table.borrow_mut().get_or_intern(value));
-        symbol.into()
+    fn from(string: &str) -> Self {
+        Self::String(string.into())
     }
 }
 
-impl From<GCHandle> for Variant {
-    fn from(handle: GCHandle) -> Self { Self::Object(handle) }
+impl From<Box<[Variant]>> for Variant {
+    fn from(items: Box<[Variant]>) -> Self {
+        Self::Tuple(GCArray::from_boxed_slice(items))
+    }
 }
 
+impl From<Function> for Variant {
+    fn from(func: Function) -> Self {
+        Self::Function(GC::allocate(func))
+    }
+}
+
+impl From<NativeFunction> for Variant {
+    fn from(func: NativeFunction) -> Self {
+        Self::NativeFunction(GC::allocate(func))
+    }
+}
+
+impl GCTrace for Variant { }
+
 
 // Not all Variants are hashable, so there is a separate type to handle that
 
@@ -205,7 +230,8 @@ impl fmt::Debug for Variant {
                 write!(fmt, "{:?})", last)
             }
             
-            Self::Object(handle) => write!(fmt, "{:?}", handle),
+            Self::Function(fun) => write!(fmt, "<{}>", fun.signature()),
+            Self::NativeFunction(fun) => write!(fmt, "<built-in {}>", fun.signature()),
         }
     }
 }
