@@ -1,10 +1,11 @@
 use crate::language::{IntType, FloatType};
 use crate::codegen::{Program, ProgramData, ChunkID, ConstID, Constant, OpCode};
 use crate::runtime::Variant;
+use crate::runtime::gc::GC;
 use crate::runtime::ops;
-use crate::runtime::types::Call;
+use crate::runtime::types::function::Call;
 use crate::runtime::strings::StringSymbol;
-use crate::runtime::module::{Module, ModuleCache, ModuleID, Access, GlobalEnv};
+use crate::runtime::module::{Module, Access, GlobalEnv};
 use crate::runtime::errors::{ExecResult, RuntimeError, ErrorKind};
 use crate::debug::traceback::CallSite;
 use crate::debug::snapshot::{VMSnapshot, VMStateSnapshot};
@@ -12,6 +13,7 @@ use crate::debug::snapshot::{VMSnapshot, VMStateSnapshot};
 
 // data used to set up a new call
 struct CallInfo {
+    nargs: usize,
     frame: usize,
     call: Call,
     site: CallSite,
@@ -27,38 +29,22 @@ enum Control {
 
 // Stack-based Virtual Machine
 #[derive(Debug)]
-pub struct VirtualMachine<'m> {
-    module_cache: &'m ModuleCache,
+pub struct VirtualMachine<'c> {
     traceback: Vec<CallSite>,
     
-    calls: Vec<VMState<'m>>,
+    calls: Vec<VMState<'c>>,
     values: ValueStack,
-    state: VMState<'m>,
+    state: VMState<'c>,
 }
 
-impl<'m> VirtualMachine<'m> {
+impl<'c> VirtualMachine<'c> {
     /// Create a new VM with the specified root module and an empty main chunk
-    pub fn new(module_cache: &'m ModuleCache, module_id: ModuleID, main_chunk: &'m [u8]) -> Self {
-        let module = module_cache.get(&module_id).expect("invalid module id");
-        
+    pub fn new(main_module: GC<Module>, main_chunk: &'c [u8]) -> Self {
         Self {
-            module_cache,
             traceback: Vec::new(),
             calls: Vec::new(),
             values: ValueStack::new(),
-            state: VMState::main_chunk(module, main_chunk),
-        }
-    }
-    
-    pub fn repl(module_cache: &'m ModuleCache, repl_env: &'m GlobalEnv, module_id: ModuleID, main_chunk: &'m [u8]) -> Self {
-        let module = module_cache.get(&module_id).expect("invalid module id");
-        
-        Self {
-            module_cache,
-            traceback: Vec::new(),
-            calls: Vec::new(),
-            values: ValueStack::new(),
-            state: VMState::main_repl(module, repl_env, main_chunk),
+            state: VMState::main_chunk(main_module, main_chunk),
         }
     }
     
@@ -67,13 +53,15 @@ impl<'m> VirtualMachine<'m> {
         Ok(())
     }
     
-    pub fn run_steps(self) -> impl Iterator<Item=ExecResult<VMSnapshot>> + 'm {
+    pub fn run_steps(self) -> impl Iterator<Item=ExecResult<VMSnapshot>> + 'c {
         VMStepper::from(self)
     }
     
     #[inline]
     fn exec_next(&mut self) -> ExecResult<bool> {
-        match self.state.exec_next(&mut self.values)? {
+        let control = self.state.exec_next(&mut self.values)?;
+        
+        match control {
             Control::Exit => return Ok(true),
             Control::Return if self.calls.is_empty() => return Ok(true),
             
@@ -81,6 +69,7 @@ impl<'m> VirtualMachine<'m> {
             Control::Return => self.return_call(),
             Control::Continue => { }
         }
+        
         Ok(false)
     }
     
@@ -88,17 +77,15 @@ impl<'m> VirtualMachine<'m> {
         self.traceback.push(call.site);
         
         match call.call {
-            Call::Native(retval) => {
-                let retval = retval?;
+            Call::Native(func) => {
+                let args = self.values.peek_many(call.nargs);
+                let retval = func.invoke(args)?;
                 self.values.truncate(call.frame.into());
                 self.values.push(retval);
                 self.traceback.pop();
             },
             
-            Call::Chunk(module_id, chunk_id) => {
-                let module = self.module_cache.get(&module_id)
-                    .expect("invalid module id");
-                
+            Call::Chunk(module, chunk_id) => {
                 let locals = LocalIndex::try_from(self.values.len() - call.frame)
                     .expect("local index overflow");
                 
@@ -180,64 +167,50 @@ macro_rules! cond_jump {
 }
 
 
-pub type LocalIndex = u16;  // TODO make this u32 and add opcodes
+pub type LocalIndex = u16;
+
 
-// stores the active chunk execution information
 #[derive(Debug)]
-struct VMState<'m> {
-    module: &'m Module,
-    globals: &'m GlobalEnv,
-    chunk: &'m [u8],
+struct VMState<'c> {
+    module: GC<Module>,
+    chunk: &'c [u8],
     chunk_id: Option<ChunkID>,
-    frame: usize,
-    locals: LocalIndex,
+    frame: usize,        // start index for locals belonging to this frame
+    locals: LocalIndex,  // local variable count
     pc: usize,
 }
 
-impl<'m> VMState<'m> {
-    fn call_frame(module: &'m Module, chunk_id: ChunkID, frame: usize, locals: LocalIndex) -> Self {
+impl<'c> VMState<'c> {
+    fn call_frame(module: GC<Module>, chunk_id: ChunkID, frame: usize, locals: LocalIndex) -> Self {
+        
+        // This hack allows us to get around the self-referentiality of storing both "module" and "chunk"
+        // in the same struct. The alternative would be to store "chunk" as an `Option<Box<[u8]>>` and call 
+        // `module.data().get_chunk()` before every single instruction, but I want to avoid the overhead of that.
+        // SAFETY: This is safe because "module" is rooted as long as this VMState is in the call stack, 
+        // and VMStates are never stored outside of a VirtualMachine's call stack.
+        let chunk: *const [u8] = module.data().get_chunk(chunk_id);
+        let chunk = unsafe { chunk.as_ref::<'c>().unwrap() };
+        
         Self {
             module,
-            globals: module.globals(),
-            chunk: module.data().get_chunk(chunk_id),
+            chunk,
             chunk_id: Some(chunk_id),
-            locals,
             frame,
+            locals,
             pc: 0,
         }
     }
     
-    fn main_chunk(module: &'m Module, chunk: &'m [u8]) -> Self {
+    fn main_chunk(module: GC<Module>, chunk: &'c [u8]) -> Self {
         Self {
             module,
-            globals: module.globals(),
             chunk,
             chunk_id: None,
-            locals: 0,
             frame: 0,
-            pc: 0,
-        }
-    }
-    
-    fn main_repl(module: &'m Module, globals: &'m GlobalEnv, chunk: &'m [u8]) -> Self {
-        Self {
-            module,
-            globals,
-            chunk,
-            chunk_id: None,
             locals: 0,
-            frame: 0,
             pc: 0,
         }
     }
-    
-    fn get_callsite(&self, offset: usize) -> CallSite {
-        CallSite::Chunk {
-            offset,
-            module_id: self.module.module_id(),
-            chunk_id: self.chunk_id,
-        }
-    }
 
     #[inline(always)]
     fn offset_pc(&self, offset: isize) -> Option<usize> {
@@ -266,6 +239,20 @@ impl<'m> VMState<'m> {
         panic!("invalid operand")
     }
 
+    // convert a local variable index to an index into the value stack
+    #[inline]
+    fn local_index(&self, local: LocalIndex) -> usize {
+        self.frame + usize::from(local)
+    }
+    
+    fn get_callsite(&self, offset: usize) -> CallSite {
+        CallSite::Chunk {
+            offset,
+            module: self.module,
+            chunk_id: self.chunk_id,
+        }
+    }
+
     #[inline(always)]
     fn exec_next(&mut self, stack: &mut ValueStack) -> ExecResult<Control> {
         let op_byte = self.chunk.get(self.pc).expect("pc out of bounds");
@@ -283,16 +270,21 @@ impl<'m> VMState<'m> {
             OpCode::Return => return Ok(Control::Return),
             
             OpCode::Call => {
-                let nargs = Self::into_usize(stack.peek().clone());
-                stack.swap_last(stack.len() - nargs - 1);
+                const SYSTEM_ARGS: usize = 2; // [ callee, nargs, ... ]
                 
-                let frame = stack.len() - nargs - 2;
+                let nargs = Self::into_usize(stack.peek().clone());
+                let call_locals = SYSTEM_ARGS + nargs;
+                
+                stack.swap_last(stack.len() - call_locals + 1);
+                
+                let frame = stack.len() - call_locals;
                 let callee = stack.peek_at(frame);
                 
                 let args = stack.peek_many(nargs);
                 let call = CallInfo {
+                    nargs,
+                    frame,
                     call: callee.invoke(args)?,
-                    frame: frame.try_into().expect("frame index overflow"),
                     site: self.get_callsite(current_offset),
                 };
                 return Ok(Control::Call(call))
@@ -313,71 +305,73 @@ impl<'m> VMState<'m> {
             
             OpCode::LoadConst => {
                 let cid = ConstID::from(data[0]);
-                let value = self.module.get_const(cid);
+                let value = Module::get_const(self.module, cid);
                 stack.push(value);
             },
             OpCode::LoadConst16 => {
                 let cid = ConstID::from(read_le_bytes!(u16, data));
-                let value = self.module.get_const(cid);
+                let value = Module::get_const(self.module, cid);
                 stack.push(value);
             },
             
             OpCode::InsertGlobal => {
                 let name = Self::into_name(stack.pop());
                 let value = stack.peek().clone();
-                self.globals.borrow_mut().create(name, Access::ReadOnly, value);
+                self.module.globals().borrow_mut().create(name, Access::ReadOnly, value);
             },
             OpCode::InsertGlobalMut => {
                 let name = Self::into_name(stack.pop());
                 let value = stack.peek().clone();
-                self.globals.borrow_mut().create(name, Access::ReadWrite, value);
+                self.module.globals().borrow_mut().create(name, Access::ReadWrite, value);
             },
             OpCode::StoreGlobal => {
                 let name = Self::into_name(stack.pop());
                 let value = stack.peek().clone();
                 
-                let mut globals = self.globals.borrow_mut();
+                let mut globals = self.module.globals().borrow_mut();
                 let store = globals.lookup_mut(&name)?;
                 *store = value;
             },
             OpCode::LoadGlobal => {
                 let value = {
                     let name = Self::into_name(stack.peek().clone());
-                    self.globals.borrow().lookup(&name)?.clone()
+                    self.module.globals().borrow().lookup(&name)?.clone()
                 };
                 stack.replace(value);
             },
             
             OpCode::InsertLocal => {
                 let value = stack.peek().clone();
-                stack.insert(self.locals.into(), value);
+                stack.insert(self.local_index(self.locals), value);
                 self.locals += 1;
             },
             OpCode::StoreLocal => {
-                let index = usize::from(data[0]);
-                stack.replace_at(self.frame + index, stack.peek().clone());
+                let index = LocalIndex::from(data[0]);
+                stack.replace_at(self.local_index(index), stack.peek().clone());
             },
             OpCode::StoreLocal16 => {
-                let index = usize::from(read_le_bytes!(u16, data));
-                stack.replace_at(self.frame + index, stack.peek().clone());
+                let index = LocalIndex::from(read_le_bytes!(u16, data));
+                stack.replace_at(self.local_index(index), stack.peek().clone());
             },
             OpCode::LoadLocal => {
-                let index = usize::from(data[0]);
-                debug_assert!(index < self.locals.into());
-                stack.push(stack.peek_at(self.frame + index).clone());
+                let index = LocalIndex::from(data[0]);
+                debug_assert!(index < self.locals);
+                stack.push(stack.peek_at(self.local_index(index)).clone());
             },
             OpCode::LoadLocal16 => {
-                let index = usize::from(read_le_bytes!(u16, data));
+                let index = LocalIndex::from(read_le_bytes!(u16, data));
                 debug_assert!(index < self.locals.into());
-                stack.push(stack.peek_at(self.frame + index).clone());
+                stack.push(stack.peek_at(self.local_index(index)).clone());
             },
             OpCode::DropLocals => {
                 let count = LocalIndex::from(data[0]);
-                let locals = self.frame + usize::from(self.locals);
-                if stack.len() == locals {
+                debug_assert!(count <= self.locals);
+                
+                let locals_end_index = self.local_index(self.locals);
+                if stack.len() == locals_end_index {
                     stack.discard(usize::from(count));
                 } else {
-                    let index = usize::from(locals) - usize::from(count);
+                    let index = locals_end_index - usize::from(count);
                     stack.discard_at(index, usize::from(count));
                 }
                 self.locals -= count;
@@ -391,21 +385,17 @@ impl<'m> VMState<'m> {
             OpCode::Tuple => {
                 let tuple_len = usize::from(data[0]);
                 
-                // peek and clone to ensure items stay rooted for GC
-                let items = stack.peek_many(tuple_len)
-                    .to_vec().into_boxed_slice();
+                let items = stack.pop_many(tuple_len).into_boxed_slice();
                 
-                let tuple = Variant::make_tuple(items);
+                let tuple = Variant::from(items);
                 stack.replace_many(tuple_len, tuple);
             },
             OpCode::TupleN => {
                 let tuple_len = Self::into_usize(stack.pop());
                 
-                // peek and clone to ensure items stay rooted for GC
-                let items = stack.peek_many(tuple_len)
-                    .to_vec().into_boxed_slice();
+                let items = stack.pop_many(tuple_len).into_boxed_slice();
                 
-                let tuple = Variant::make_tuple(items);
+                let tuple = Variant::from(items);
                 stack.replace_many(tuple_len, tuple);
             },
             
@@ -610,7 +600,7 @@ impl From<&VMState<'_>> for VMStateSnapshot {
             ));
         
         Self {
-            module_id: state.module.module_id(),
+            module: state.module.name().to_string(),
             chunk_id: state.chunk_id,
             frame: state.frame,
             locals: state.locals,
